/*
 * Copyright (c) "Neo4j"
 * Neo4j Sweden AB [http://neo4j.com]
 *
 * This file is part of Neo4j.
 *
 * Neo4j is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.neo4j.kernel.impl.util;

import java.util.ArrayList;

import org.neo4j.graphdb.Label;
import org.neo4j.graphdb.Node;
import org.neo4j.graphdb.NotFoundException;
import org.neo4j.kernel.impl.core.NodeProxy;
import org.neo4j.kernel.impl.store.InvalidRecordException;
import org.neo4j.values.AnyValueWriter;
import org.neo4j.values.storable.TextArray;
import org.neo4j.values.storable.Values;
import org.neo4j.values.virtual.MapValue;
import org.neo4j.values.virtual.NodeValue;
import org.neo4j.values.virtual.VirtualValues;

public class NodeProxyWrappingNodeValue extends NodeValue
{
    private final Node node;
    private volatile TextArray labels;
    private volatile MapValue properties;

    NodeProxyWrappingNodeValue( Node node )
    {
        super( node.getId() );
        this.node = node;
    }

    public Node nodeProxy()
    {
        return node;
    }

    @Override
    public <E extends Exception> void writeTo( AnyValueWriter<E> writer ) throws E
    {
        TextArray l;
        MapValue p;
        try
        {
            l = labels();
            p = properties();
        }
        catch ( ReadAndDeleteTransactionConflictException e )
        {
            if ( !e.wasDeletedInThisTransaction() )
            {
                throw e;
            }
            // If it isn't a transient error then the node was deleted in the current transaction and we should write an 'empty' node.
            l = Values.stringArray();
            p = VirtualValues.EMPTY_MAP;
        }

        if ( id() < 0 )
        {
            writer.writeVirtualNodeHack( node );
        }

        writer.writeNode( node.getId(), l, p );
    }

    @Override
    public TextArray labels()
    {
        TextArray l = labels;
        if ( l == null )
        {
            try
            {
                synchronized ( this )
                {
                    l = labels;
                    if ( l == null )
                    {
                        ArrayList<String> ls = new ArrayList<>();
                        for ( Label label : node.getLabels() )
                        {
                            ls.add( label.name() );
                        }
                        l = labels = Values.stringArray( ls.toArray( new String[0] ) );
                    }
                }
            }
            catch ( NotFoundException | IllegalStateException | InvalidRecordException e )
            {
                throw new ReadAndDeleteTransactionConflictException( NodeProxy.isDeletedInCurrentTransaction( node ), e );
            }
        }
        return l;
    }

    @Override
    public MapValue properties()
    {
        MapValue m = properties;
        if ( m == null )
        {
            try
            {
                synchronized ( this )
                {
                    m = properties;
                    if ( m == null )
                    {
                        m = properties = ValueUtils.asMapValue( node.getAllProperties() );
                    }
                }
            }
            catch ( NotFoundException | IllegalStateException | InvalidRecordException e )
            {
                throw new ReadAndDeleteTransactionConflictException( NodeProxy.isDeletedInCurrentTransaction( node ), e );
            }
        }
        return m;
    }
}
