/*
 * Copyright (c) "Neo4j"
 * Neo4j Sweden AB [http://neo4j.com]
 *
 * This file is part of Neo4j.
 *
 * Neo4j is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.neo4j.server.rest.transactional;

import org.junit.Test;

import java.io.ByteArrayInputStream;
import java.util.Iterator;

import org.neo4j.kernel.api.exceptions.Status;
import org.neo4j.server.rest.transactional.error.Neo4jError;
import org.neo4j.string.UTF8;

import static java.util.Arrays.asList;
import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.equalTo;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;
import static org.neo4j.helpers.collection.MapUtil.map;
import static org.neo4j.server.rest.domain.JsonHelper.createJsonFrom;

public class StatementDeserializerTest
{
    @Test
    public void shouldDeserializeSingleStatement()
    {
        // Given
        String json = createJsonFrom( map( "statements", asList( map( "statement", "Blah blah", "parameters", map( "one", 12 ) ) ) ) );

        // When
        StatementDeserializer de = new StatementDeserializer( new ByteArrayInputStream( UTF8.encode( json ) ) );

        // Then
        assertThat( de.hasNext(), equalTo( true ) );
        Statement stmt = de.next();

        assertThat( stmt.statement(), equalTo( "Blah blah" ) );
        assertThat( stmt.parameters(), equalTo( map( "one", 12 ) ) );

        assertThat( de.hasNext(), equalTo( false ) );
    }

    @Test
    public void shouldRejectMapWithADifferentFieldBeforeStatement()
    {
        // NOTE: We don't really want this behaviour, but it's a symptom of keeping
        // streaming behaviour while moving the statement list into a map.

        String json = "{ \"timeout\" : 200, \"statements\" : [ { \"statement\" : \"ignored\", \"parameters\" : {}} ] }";

        assertYieldsErrors( json, new Neo4jError( Status.Request.InvalidFormat, new DeserializationException(
                "Unable to deserialize request. Expected first field to be 'statements', but was 'timeout'." ) ) );
    }

    @Test
    public void shouldTotallyIgnoreInvalidJsonAfterStatementArrayHasFinished()
    {
        // NOTE: We don't really want this behaviour, but it's a symptom of keeping
        // streaming behaviour while moving the statement list into a map.

        // Given
        String json =  "{ \"statements\" : [ { \"statement\" : \"Blah blah\", \"parameters\" : {\"one\" : 12}} ] " +
                "totally invalid json is totally ignored";

        // When
        StatementDeserializer de = new StatementDeserializer( new ByteArrayInputStream( UTF8.encode( json ) ) );

        // Then
        assertThat( de.hasNext(), equalTo( true ) );
        Statement stmt = de.next();

        assertThat( stmt.statement(), equalTo( "Blah blah" ) );

        assertThat( de.hasNext(), equalTo( false ) );
    }

    @Test
    public void shouldIgnoreUnknownFields()
    {
        // Given
        String json =  "{ \"statements\" : [ { \"a\" : \"\", \"b\" : { \"k\":1 }, \"statement\" : \"blah\" } ] }";

        // When
        StatementDeserializer de = new StatementDeserializer( new ByteArrayInputStream( UTF8.encode( json ) ) );

        // Then
        assertThat( de.hasNext(), equalTo( true ) );

        assertThat( de.next().statement(), equalTo( "blah" ) );
        assertThat( de.hasNext(), equalTo( false ) );
    }

    @Test
    public void shouldTakeParametersBeforeStatement()
    {
        // Given
        String json =  "{ \"statements\" : [ { \"a\" : \"\", \"parameters\" : { \"k\":1 }, \"statement\" : \"blah\"}]}";

        // When
        StatementDeserializer de = new StatementDeserializer( new ByteArrayInputStream( UTF8.encode( json ) ) );

        // Then
        assertThat( de.hasNext(), equalTo( true ) );

        Statement stmt = de.next();
        assertThat( stmt.statement(), equalTo( "blah" ) );
        assertThat( stmt.parameters(), equalTo( map("k", 1) ) );

        assertThat( de.hasNext(), equalTo( false ) );
    }

    @Test
    public void shouldTreatEmptyInputStreamAsEmptyStatementList()
    {
        // Given
        byte[] json = new byte[0];

        // When
        StatementDeserializer de = new StatementDeserializer( new ByteArrayInputStream( json ) );

        // Then
        assertFalse( de.hasNext() );
        assertFalse( de.errors().hasNext() );
    }

    @Test
    @SuppressWarnings( "unchecked" )
    public void shouldDeserializeMultipleStatements()
    {
        // Given
        String json = createJsonFrom( map( "statements", asList(
                map( "statement", "Blah blah", "parameters", map( "one", 12 ) ),
                map( "statement", "Blah bluh", "parameters", map( "asd", asList( "one, two" ) ) ) ) ) );

        // When
        StatementDeserializer de = new StatementDeserializer( new ByteArrayInputStream( UTF8.encode( json ) ) );

        // Then
        assertThat( de.hasNext(), equalTo( true ) );
        Statement stmt = de.next();

        assertThat( stmt.statement(), equalTo( "Blah blah" ) );
        assertThat( stmt.parameters(), equalTo( map( "one", 12 ) ) );

        assertThat( de.hasNext(), equalTo( true ) );
        Statement stmt2 = de.next();

        assertThat( stmt2.statement(), equalTo( "Blah bluh" ) );
        assertThat( stmt2.parameters(), equalTo( map( "asd", asList( "one, two" ) ) ) );

        assertThat( de.hasNext(), equalTo( false ) );
    }

    @Test
    public void shouldNotThrowButReportErrorOnInvalidInput()
    {
        assertYieldsErrors( "{}",
                new Neo4jError( Status.Request.InvalidFormat, new DeserializationException( "Unable to " +
                        "deserialize request. " +
                        "Expected [START_OBJECT, FIELD_NAME, START_ARRAY], " +
                        "found [START_OBJECT, END_OBJECT, null]." ) ) );

        assertYieldsErrors( "{ \"statements\":\"WAIT WAT A STRING NOO11!\" }",
                new Neo4jError( Status.Request.InvalidFormat, new DeserializationException( "Unable to " +
                        "deserialize request. Expected [START_OBJECT, FIELD_NAME, START_ARRAY], found [START_OBJECT, " +
                        "FIELD_NAME, VALUE_STRING]." ) ) );

        assertYieldsErrors( "[{]}",
                new Neo4jError( Status.Request.InvalidFormat,
                        new DeserializationException( "Unable to deserialize request: Unexpected close marker ']': " +
                                "expected '}' " +
                                "(for Object starting at [Source: (ByteArrayInputStream); line: 1, column: 2])\n " +
                                "at [Source: (ByteArrayInputStream); line: 1, column: 4]" ) ) );

        assertYieldsErrors( "{ \"statements\" : \"ITS A STRING\" }",
                new Neo4jError( Status.Request.InvalidFormat,
                        new DeserializationException( "Unable to deserialize request. " +
                                "Expected [START_OBJECT, FIELD_NAME, START_ARRAY], " +
                                "found [START_OBJECT, FIELD_NAME, VALUE_STRING]." ) ) );

        assertYieldsErrors( "{ \"statements\" : [ { \"statement\" : [\"dd\"] } ] }",
                new Neo4jError( Status.Request.InvalidFormat,
                        new DeserializationException( "Unable to deserialize request: Cannot deserialize value of type" +
                                " `java.lang.String` from Array value (token `JsonToken.START_ARRAY`)\n at [Source: (ByteArrayInputStream); line: 1, " +
                                "column: 36]" ) ) );

        assertYieldsErrors( "{ \"statements\" : [ { \"statement\" : \"stmt\", \"parameters\" : [\"AN ARRAY!!\"] } ] }",
                new Neo4jError( Status.Request.InvalidFormat,
                        new DeserializationException( "Unable to deserialize request: Cannot deserialize value of type" +
                                " `java.util.LinkedHashMap<java.lang.Object,java.lang.Object>` from Array value (token `JsonToken.START_ARRAY`)\n at " +
                                "[Source: (ByteArrayInputStream); line: 1, column: 59]" ) ) );
    }

    private void assertYieldsErrors( String json, Neo4jError... expectedErrors )
    {
        StatementDeserializer de = new StatementDeserializer( new ByteArrayInputStream( UTF8.encode( json ) ) );
        while ( de.hasNext() )
        {
            de.next();
        }

        Iterator<Neo4jError> actual = de.errors();
        Iterator<Neo4jError> expected = asList( expectedErrors ).iterator();
        while ( actual.hasNext() )
        {
            assertTrue( expected.hasNext() );
            Neo4jError error = actual.next();
            Neo4jError expectedError = expected.next();

            assertThat( error.getMessage(), equalTo( expectedError.getMessage() ) );
            assertThat( error.status(), equalTo( expectedError.status() ) );
        }

        assertFalse( expected.hasNext() );
    }

}
